;**********************************************************
;*                  SER_I2C.ASM  V1.0                     *
;* Konverter von seriell (RS232) nach I2C                 *
;* Mikrocontroller: AT89C2051, f=11,0592 MHz              *
;* Serielle Uebertragung der I2C-Bus-Daten zum Mikro-     *
;* controller                                             *
;* Ser. Format: 8N1, 9600 oder 1200 Baud, kein Handshake  *
;**********************************************************
;* World Of Electronics                                   *
;* http://www.woe.de.vu                                   *
;* woe@topmail.de                                         *
;* (c) WOE, 1999                                          *
;**********************************************************

; P3.  7       5      3
;     SDA     SCL   Jumper


sda equ p3.7
scl equ p3.5
jumper equ p3.3        ; Jumper offen: 9600 Baud
                       ; Jumper an Masse: 1200 Baud

i2c_ack_in equ 0       ; vom Slave gesendetes Acknowledge-Bit
i2c_ack_out equ 1      ; an den Slave zu sendendes Acknowledge-Bit
                       ; jeweils 0=Ack, 1=kein Ack

readmode     equ 2     ; 1=Readmode, 0=Writemode
hex_valid    equ 3     ; 1=gueltiges Hexzeichen, 0=ungueltig
bin_valid    equ 4     ; 1=gueltiges Binaerzeichen, 0=ungueltig


       setb i2c_ack_out
       setb i2c_ack_in

       jnb jumper,b1200
b9600:
       mov th1, #0fdh   ; 9600 Baud
       mov tl1, #0fdh
       jmp b_end
b1200:
       mov th1, #0e8h   ; 1200 Baud
       mov tl1, #0e8h
b_end:   
       mov tmod, #20h
       mov scon, #52h
       setb tr1
       mov a, sbuf
       clr ri

wdh:

wait_1:
       jnb ri,wait_1
       mov a,sbuf
       clr ri
       cjne a,#'(',wait_1    ;  ( -> Startbedingung

start:                       ; Startbedingung
       call i2c_start

p1_w1:
       jnb ri,p1_w1
       mov a,sbuf
       clr ri
       cjne a,#')',no_stop
       jmp stopbed
no_stop:
       call check_char
       jnb hex_valid,p1_w1

       swap a
       mov r7,a

p1_w2:
       jnb ri,p1_w2
       mov a,sbuf
       clr ri
       cjne a,#')',no_stop2
       jmp stopbed
no_stop2:
       call check_char
       jnb hex_valid,p1_w2

       add a,r7
       mov c,acc.0       
       mov readmode,c
       call i2c_send
       call send_status
       jb readmode,read

write:
w_w1:
       jnb ri,w_w1
       mov a,sbuf
       clr ri
       cjne a,#')',w_no_stop
       jmp stopbed
w_no_stop:
       cjne a,#'(',w_no_start
       jmp start
w_no_start:
       call check_char
       jnb hex_valid,w_w1

       swap a
       mov r7,a

w_w2:
       jnb ri,w_w2
       mov a,sbuf
       clr ri
       cjne a,#')',w_no_stop2
       jmp stopbed
w_no_stop2:
       cjne a,#'(',w_no_start2
       jmp start
w_no_start2:
       call check_char
       jnb hex_valid,w_w2

       add a,r7
       call i2c_send
       call send_status
       
       jmp write



read:
r_w1:
       jnb ri,r_w1
       mov a,sbuf
       clr ri
       cjne a,#')',r_no_stop
       jmp stopbed
r_no_stop:
       cjne a,#'(',r_no_start
       jmp start
r_no_start:
       call check_char
       jnb bin_valid,r_w1
       
       mov c,acc.0
       mov i2c_ack_out,c
       cpl i2c_ack_out
       call i2c_read
       call send_rec_byte
       jmp read



stopbed:                     ; Stopbedingung
       call i2c_stop
       jmp wdh



send_status:
       jb i2c_ack_in,no_ack
       mov a,#'1'
       jmp send_out
no_ack:
       mov a,#'0'
send_out:
wait_ti:
       jnb ti,wait_ti
       clr ti
       mov sbuf,a
ret



send_rec_byte:
       mov b,a
       swap a
       anl a,#0fh
       call conv2ascii
s_wait_ti1:
       jnb ti,s_wait_ti1
       clr ti
       mov sbuf,a

       mov a,b
       anl a,#0fh
       call conv2ascii
s_wait_ti2:
       jnb ti,s_wait_ti2
       clr ti
       mov sbuf,a
       ret




check_char:             ; empfangenes Zeichen ueberpruefen
       clr hex_valid
       clr bin_valid

       cjne a,#48,n1
n1:
       jc ignore
       cjne a,#58,n2
n2:
       jc hex_ok1
       cjne a,#65,n3
n3:
       jc ignore
       cjne a,#71,n4
n4:
       jc hex_ok2
       cjne a,#97,n5
n5:
       jc ignore
       cjne a,#103,n6
n6:
       jc hex_ok3
       jmp ignore


ignore:            ; ungueltiges Zeichen
       ret

hex_ok1:           ; 0..9
       clr c
       subb a,#48
       setb hex_valid
       cjne a,#0,h1
       jmp bin_ok
h1:
       cjne a,#1,h2
bin_ok:
       setb bin_valid
h2:
       ret

hex_ok2:           ; A..F
       clr c
       subb a,#55
       setb hex_valid
       ret

hex_ok3:           ; a..f
       clr c
       subb a,#87
       setb hex_valid
       ret



conv2ascii:
       cjne a,#0ah,co1
co1:
       jnc z_hex
z_dez:
       add a,#48
       jmp conv_end
z_hex:
       add a,#55
conv_end:
       ret




i2c_start:                 ; I2C Startbedingung
       setb sda
       nop
       setb scl
       call i2c_wait

       clr sda             ; Data auf 0 (Start)
       nop
       clr scl             ; Clock auf 0
       ret


i2c_stop:                  ; I2C Stopbedingung
       setb scl            ; Clock auf 1
       nop
       setb sda            ; Data auf 1 (Stop)
       ret



i2c_send:                  ; 1 Datenbyte (Akku) ueber I2C senden
       mov r3,#8
i2c_sloop:
       rlc a
       mov sda,c
       nop
       setb scl
       call i2c_wait
       clr scl
       djnz r3,i2c_sloop

       nop
       setb sda
       nop
       setb scl
       call i2c_wait           
       mov c,sda            ; Acknowledge-Bit lesen
       mov i2c_ack_in,c
       clr scl
       nop
       clr sda
       ret



i2c_read:                   ; 1 Datenbyte ueber I2C empfangen (Akku)
       mov r3,#8
       clr a
       setb sda
i2c_rloop:
       call i2c_wait
       setb scl
       nop
       mov c,sda
       rlc a
       nop
       clr scl
       djnz r3,i2c_rloop

       mov c,i2c_ack_out   ; Acknowledge-Bit senden
       mov sda,c
       nop
       setb scl
       call i2c_wait
       clr scl
       nop
       clr sda
       ret


i2c_wait:               ; Verzoegerung fuer I2C (CALL + RET -> 4us bei 12MHz)
       ret


       END
